/*
 * Implement the fiwalk plugin API
 */

// needed or else glob callback doesn't work properly
#define __DARWIN_64_BIT_INO_T 1		

#include "plugin.h"
#include "config.h"
#include "arff.h"

#include <string.h>
#include <stdlib.h>
#include <regex.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <glob.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <algorithm>
#include <string>
#include <iostream>
#include <string>
#include <vector>

using namespace std;

class plugins {
public:;
    string pattern;		// what we want
    string method;
    string path;
};

vector<class plugins> plugin_list;

static bool all_whitespace(const char *buf)
{
    while(*buf){
	if(!isspace(*buf)) return false;
	buf++;
    }
    return true;
}

static const char    *mydirent_fname = 0;
static struct dirent mydirent;
static void *gl_opendir(const char *name) {
    return (void *)&mydirent;
}
static struct dirent *gl_readdir(void *d)
{
    mydirent.d_reclen = sizeof(mydirent);
    if(++mydirent.d_ino>1) return NULL;
    mydirent.d_type   = DT_REG;
    strcpy(mydirent.d_name,mydirent_fname);
#ifdef HAVE_STRUCT_DIRENT_D_NAME
    mydirent.d_namlen = strlen(mydirent.d_name);
#endif
    return &mydirent;
}

static void gl_closedir(void *v)
{
}

static int gl_stat(const char *name,struct stat *st){
    memset(st,0,sizeof(*st));
    st->st_nlink = 1;
    st->st_mode = S_IFREG;			// fake a regular file
    return 0;
}

static int gl_error(const char *path,int code)
{
    return 0;
}

bool glob_match(const char *pattern,const char *fname)
{
    memset(&mydirent,0,sizeof(mydirent));
    mydirent_fname = fname;

    glob_t g;
    memset(&g,0,sizeof(g));
    g.gl_opendir = gl_opendir;
    g.gl_readdir = gl_readdir;
    g.gl_closedir = gl_closedir;
    g.gl_lstat  = gl_stat;
    g.gl_stat   = gl_stat;
    
    int r = glob(pattern,GLOB_ALTDIRFUNC,gl_error,&g);
    if(r==GLOB_NOMATCH) return false;
    int ret = true;
#ifdef HAVE_STRUCT_GLOB_T_GL_MATCHC
    ret = g.gl_matchc >0 ? true : fnalse;
#endif
    globfree(&g);
    return ret;
}

/** Return TRUE if the FNAME requires plugin processing */
const plugins *current_plugin = 0;
bool plugin_match(const char *fname)
{
    for(vector<class plugins>::const_iterator i = plugin_list.begin();
	i != plugin_list.end();
	i++){
	if( glob_match((*i).pattern.c_str(),fname)){
	    current_plugin = &(*i);
	    return true;
	}
    }
    return false;
}

void plugin_process(const char *fname,class arff *ar)
{
    static bool first = true;
    static regex_t ncv;
    if(first){
	if(regcomp(&ncv,"([-a-zA-Z0-9_]+): (.*)",REG_EXTENDED)) err(1,"regcomp");
	first = 0;
    }


    if(current_plugin->method=="dgi"){
	string cmd = current_plugin->path + " " + fname;
	FILE *f = popen(cmd.c_str(),"r");
	if(!f) err(1,"fopen: %s",cmd.c_str());
	char linebuf[1024];
	while (fgets(linebuf,sizeof(linebuf),f)){
	    char *cc = rindex(linebuf,'\n');
	    if(cc) *cc = '\000';
	    
	    /* process name: value pairs */
	    regmatch_t pmatch[4];
	    memset(pmatch,0,sizeof(pmatch));
	    if(regexec(&ncv,linebuf,4,pmatch,0)) errx(1,"plugin %s returned: '%s'",
						    current_plugin->path.c_str(),linebuf);
	    linebuf[pmatch[1].rm_eo] = 0;
	    linebuf[pmatch[2].rm_eo] = 0;
	    const char *name = linebuf+pmatch[1].rm_so;
	    const char *value = linebuf+pmatch[2].rm_so;
	    if(ar){
		ar->add_attribute(name,arff::STRING);
		ar->add_value(name,value);
	    }
	}
	pclose(f);
	return;
    }
}


void config_read(const char *fname)
{
    /* make sure the glob function works */
    assert(glob_match("*.jpeg","file.jpeg")==true);
    assert(glob_match("*.jpg","file.jpeg")==false);

    regex_t r;
    if(regcomp(&r,"([^ \t]+)[ \t]+([^ \t]+)[ \t]+([^ \t\r\n]+)",REG_EXTENDED)) err(1,"regcomp"); // No support for \s
    FILE *f = fopen(fname,"r");
    if(!f) err(1,fname);
    char linebuf[1024];
    int linenumber = 0;
    while(fgets(linebuf,sizeof(linebuf),f)){
	linenumber++;
	char *cc = strchr(linebuf,'#');
	if(cc) *cc = 0;			// terminate #'s

	/* if the line all whitespace ignore it */
	if(all_whitespace(linebuf)) continue;

	/* parse the line */
	regmatch_t pmatch[10];
	memset(pmatch,0,sizeof(pmatch));
	int res = regexec(&r,linebuf,10,pmatch,0);
	if(res){
	    fprintf(stderr,"Error in configuration file line %d: %s\n",linenumber,linebuf);
	    exit(1);
	}
	class plugins plug;
	linebuf[pmatch[1].rm_eo] = 0;
	plug.pattern = linebuf+pmatch[1].rm_so;

	linebuf[pmatch[2].rm_eo] = 0;
	plug.method = linebuf+pmatch[2].rm_so;

	linebuf[pmatch[3].rm_eo] = 0;
	plug.path = linebuf+pmatch[3].rm_so;
	plugin_list.push_back(plug);
    }
    fclose(f);
    regfree(&r);
}
